@@ -140,10 +140,70 @@ struct ParsedConfigFile {
 }
 
 impl ParsedConfigFile {
-    fn parse(s: &str) -> Self {
+    fn parse(s: &str, cwd: Option<&Path>) -> Self {
         let mut options = ConfigMap::new();
         let mut groups = vec![];
 
+        Self::parse_impl(s, cwd, &mut options, &mut groups);
+
+        Self { options, groups }
+    }
+
+    fn do_include(
+        pattern: &str,
+        cwd: Option<&Path>,
+        options: &mut ConfigMap,
+        groups: &mut Vec<MatchGroup>,
+    ) {
+        match filenamegen::Glob::new(&pattern) {
+            Ok(g) => {
+                match cwd
+                    .as_ref()
+                    .map(|p| p.to_path_buf())
+                    .or_else(|| std::env::current_dir().ok())
+                {
+                    Some(cwd) => {
+                        for path in g.walk(&cwd) {
+                            let path = if path.is_absolute() {
+                                path
+                            } else {
+                                cwd.join(path)
+                            };
+                            match std::fs::read_to_string(&path) {
+                                Ok(data) => {
+                                    Self::parse_impl(&data, Some(&cwd), options, groups);
+                                }
+                                Err(err) => {
+                                    log::error!(
+                                        "error expanding `Include {}`: unable to open {}: {:#}",
+                                        pattern,
+                                        path.display(),
+                                        err
+                                    );
+                                }
+                            }
+                        }
+                    }
+                    None => {
+                        log::error!(
+                            "error expanding `Include {}`: unable to determine cwd",
+                            pattern
+                        );
+                    }
+                }
+            }
+            Err(err) => {
+                log::error!("error expanding `Include {}`: {:#}", pattern, err);
+            }
+        }
+    }
+
+    fn parse_impl(
+        s: &str,
+        cwd: Option<&Path>,
+        options: &mut ConfigMap,
+        groups: &mut Vec<MatchGroup>,
+    ) {
         for line in s.lines() {
             let line = line.trim();
             if line.is_empty() || line.starts_with('#') {
@@ -176,9 +236,26 @@ impl ParsedConfigFile {
                     }
                     patterns
                 }
+                fn parse_whitespace_pattern_list(v: &str) -> Vec<Pattern> {
+                    let mut patterns = vec![];
+                    for p in v.split_ascii_whitespace() {
+                        let p = p.trim();
+                        if p.starts_with('!') {
+                            patterns.push(Pattern::new(&p[1..], true));
+                        } else {
+                            patterns.push(Pattern::new(p, false));
+                        }
+                    }
+                    patterns
+                }
+
+                if k == "include" {
+                    Self::do_include(v, cwd, options, groups);
+                    continue;
+                }
 
                 if k == "host" {
-                    let patterns = parse_pattern_list(v);
+                    let patterns = parse_whitespace_pattern_list(v);
                     groups.push(MatchGroup {
                         criteria: vec![Criteria::Host(patterns)],
                         options: ConfigMap::new(),
@@ -241,20 +318,28 @@ impl ParsedConfigFile {
                     continue;
                 }
 
-                if let Some(group) = groups.last_mut() {
-                    group
-                        .options
-                        .entry(k.to_string())
-                        .or_insert_with(|| v.to_string());
-                } else {
+                fn add_option(options: &mut ConfigMap, k: String, v: &str) {
+                    // first option wins in ssh_config, except for identityfile
+                    // which explicitly allows multiple entries to combine together
+                    let is_identity_file = k == "identityfile";
                     options
-                        .entry(k.to_string())
+                        .entry(k)
+                        .and_modify(|e| {
+                            if is_identity_file {
+                                e.push(' ');
+                                e.push_str(v);
+                            }
+                        })
                         .or_insert_with(|| v.to_string());
                 }
+
+                if let Some(group) = groups.last_mut() {
+                    add_option(&mut group.options, k, v);
+                } else {
+                    add_option(options, k, v);
+                }
             }
         }
-
-        Self { options, groups }
     }
 
     /// Apply configuration values that match the specified hostname to target,
@@ -337,14 +422,15 @@ impl Config {
     /// and add that to the list of configs.
     pub fn add_config_string(&mut self, config_string: &str) {
         self.config_files
-            .push(ParsedConfigFile::parse(config_string));
+            .push(ParsedConfigFile::parse(config_string, None));
     }
 
     /// Open `path`, read its contents and parse it as an `ssh_config` file,
     /// adding that to the list of configs
     pub fn add_config_file<P: AsRef<Path>>(&mut self, path: P) {
-        if let Ok(data) = std::fs::read_to_string(path) {
-            self.config_files.push(ParsedConfigFile::parse(&data));
+        if let Ok(data) = std::fs::read_to_string(path.as_ref()) {
+            self.config_files
+                .push(ParsedConfigFile::parse(&data, path.as_ref().parent()));
         }
     }
 
@@ -354,7 +440,10 @@ impl Config {
         if let Some(home) = dirs_next::home_dir() {
             self.add_config_file(home.join(".ssh").join("config"));
         }
-        self.add_config_file("/etc/ssh/config");
+        self.add_config_file("/etc/ssh/ssh_config");
+        if let Ok(sysdrive) = std::env::var("SystemDrive") {
+            self.add_config_file(format!("{}/ProgramData/ssh/ssh_config", sysdrive));
+        }
     }
 
     fn resolve_local_user(&self) -> String {
@@ -501,13 +590,22 @@ impl Config {
         for &t in tokens {
             if let Some(v) = self.tokens.get(t) {
                 *value = value.replace(t, v);
+            } else if t == "%u" {
+                *value = value.replace(t, &self.resolve_local_user());
             } else if t == "%d" {
                 if let Some(home) = self.resolve_home() {
-                    if value.starts_with("~/") {
-                        value.replace_range(0..1, &home);
-                    } else {
-                        *value = value.replace(t, &home);
+                    let mut items = value
+                        .split_whitespace()
+                        .map(|s| s.to_string())
+                        .collect::<Vec<String>>();
+                    for item in &mut items {
+                        if item.starts_with("~/") {
+                            item.replace_range(0..1, &home);
+                        } else {
+                            *item = item.replace(t, &home);
+                        }
                     }
+                    *value = items.join(" ");
                 }
             }
         }
@@ -620,6 +718,40 @@ Config {
         );
     }
 
+    #[test]
+    fn multiple_identityfile() {
+        let mut config = Config::new();
+
+        let mut fake_env = ConfigMap::new();
+        fake_env.insert("HOME".to_string(), "/home/me".to_string());
+        fake_env.insert("USER".to_string(), "me".to_string());
+        config.assign_environment(fake_env);
+
+        config.add_config_string(
+            r#"
+        Host foo
+            HostName 10.0.0.1
+            User foo
+            IdentityFile "~/.ssh/id_pub.dsa"
+            IdentityFile "~/.ssh/id_pub.rsa"
+            "#,
+        );
+
+        let opts = config.for_host("foo");
+        snapshot!(
+            opts,
+            r#"
+{
+    "hostname": "10.0.0.1",
+    "identityfile": "/home/me/.ssh/id_pub.dsa /home/me/.ssh/id_pub.rsa",
+    "port": "22",
+    "user": "foo",
+    "userknownhostsfile": "/home/me/.ssh/known_hosts /home/me/.ssh/known_hosts2",
+}
+"#
+        );
+    }
+
     #[test]
     fn sub_tilde() {
         let mut config = Config::new();
@@ -906,11 +1038,11 @@ Config {
         Something first
         # the prior Something takes precedence
         Something ignored
-        Host 192.168.1.8,wopr
+        Host 192.168.1.8 wopr
             FowardAgent yes
             IdentityFile "%d/.ssh/id_pub.dsa"
 
-        Host !a.b,*.b
+        Host !a.b *.b
             ForwardAgent no
             IdentityAgent "${HOME}/.ssh/agent"
 
